/* vi:set wm=0 ai tabstop=4 shiftwidth=4 sm: */

/*
 * sctl - Bearcat scanner control program
 * Copyright (C) 2001 Ian Johnston <reaper@muppetlabs.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* unchecked limit, talkgroup, talkgroup_all */

/***********************************************************************
 * commands for the scanner go in here.
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include "config.h"
#include "serial.h"
#include "commands.h"
#include "utils.h"

/***********************************************************************
 * data initializations
 ***********************************************************************/

char *keypad_nums[] =
{
	K_0,
	K_1,
	K_2,
	K_3,
	K_4,
	K_5,
	K_6,
	K_7,
	K_8,
	K_9
};

/* Mode descriptions */
char *mode_desc[] =
{
	"Conventional channel scan mode",		/* 0 */
	"Manual mode",							/* 1 */
	"Limit search mode",					/* 2 */
	"Limit search hold mode",				/* 3 */
	"Weather scan mode",					/* 4 */
	"Weather scan hold mode",				/* 5 */
	"Service scan mode",					/* 6 */
	"Service scan hold mode",				/* 7 */
	"Motorola program trunk frequency mode", /* 8 */
	"Edacs program trunk frequency mode",	/* 9 */
	"Trunk system program mode",			/* 10 */
	"Motorola id search mode",				/* 11 */
	"Motorola id search hold mode",			/* 12 */
	"Motorola id scan mode",				/* 13 */
	"Motorola id manual mode",				/* 14 */
	"Motorola id lockout review mode",		/* 15 */
	"Motorola search control channel mode",	/* 16 */
	"Edacs id search mode",					/* 17 */
	"Edacs id search hold mode",			/* 18 */
	"Edacs id scan mode",					/* 19 */
	"Edacs id manual mode",					/* 20 */
	"Edacs id lockout review mode",			/* 21 */
	"Edacs search control channel mode",	/* 22 */
	"Rotary tuned frequency mode",			/* 23 */
	"Program CTCSS mode",					/* 24 */
	"Weather alert mode",					/* 25 */
	"Frequency send mode",					/* 26 */
	"Auto store mode",						/* 27 */
	"RF tune mode",							/* 28 */
	"Unknown mode"							/* 29 */
};

/***********************************************************************
 * RF attenuation
 ***********************************************************************/
int att(int fd, int att_var)
{
	char answer[100];

	bzero(answer, 100);

	if (att_var == -1)
	{
		do_command(fd, "AT", answer);

		if (strncmp(answer, "ATN", 3) == 0)
		{
			puts("RF attenuation is ON.");
			return(1);
		}
		else if (strncmp(answer, "ATF", 3) == 0)
		{
			puts("RF attenuation is OFF.");
			return(1);
		}
		else
			u_err("AT", answer);
	}
	else if (att_var == 0)
	{
		do_command(fd, "ATF", answer);

		if (strncmp(answer, "OK", 2) == 0)
		{
			puts("Turned RF attenuation OFF.");
			return(1);
		}
		else
			u_err("ATF", answer);
	}
	else if (att_var == 1)
	{
		do_command(fd, "ATN", answer);

		if (strncmp(answer, "OK", 2) == 0)
		{
			puts("Turned RF attenuation ON.");
			return(1);
		}
		else
			u_err("ATN", answer);
	}

	return(1);
}


/***********************************************************************
 * data skip
 ***********************************************************************/
int data(int fd, int data_var)
{
	char answer[100];

	bzero(answer, 100);

	if (data_var == -1)
	{
		do_command(fd, "DS", answer);

		if (strncmp(answer, "DSN", 3) == 0)
		{
			puts("Data skip is ON.");
			return(1);
		}
		else if (strncmp(answer, "DSF", 3) == 0)
		{
			puts("Data skip is OFF.");
			return(1);
		}
		else
			u_err("DS", answer);
	}
	else if (data_var == 0)
	{
		do_command(fd, "DSF", answer);

		if (strncmp(answer, "OK", 2) == 0)
		{
			puts("Turned data skip OFF.");
			return(1);
		}
		else
			u_err("DSF", answer);
	}
	else if (data_var == 1)
	{
		do_command(fd, "DSN", answer);

		if (strncmp(answer, "OK", 2) == 0)
		{
			puts("Turned data skip ON.");
			return(1);
		}
		else
			u_err("DSN", answer);
	}

	return(1);
}


/***********************************************************************
 * Delay
 ***********************************************************************/
int delay(int fd, int delay_var)
{
	char answer[100];

	bzero(answer, 100);

	if (delay_var == -1)
	{
		do_command(fd, "DL", answer);

		if (strncmp(answer, "DLN", 3) == 0)
		{
			puts("Delay function is ON for this channel.");
			return(1);
		}
		else if (strncmp(answer, "DLF", 3) == 0)
		{
			puts("Delay function is OFF for this channel.");
			return(1);
		}
		else
			u_err("DL", answer);
	}
	else if (delay_var == 0)
	{
		do_command(fd, "DLF", answer);

		if (strncmp(answer, "OK", 2) == 0)
		{
			puts("Set delay OFF for this channel.");
			return(1);
		}
		else
			u_err("DLF", answer);
	}
	else if (delay_var == 1)
	{
		do_command(fd, "DLN", answer);

		if (strncmp(answer, "OK", 2) == 0)
		{
			puts("Set delay ON for this channel.");
			return(1);
		}
		else
			u_err("DLN", answer);
	}

	return(1);
}


/***********************************************************************
 * Backup a bank to file
 ***********************************************************************/
int dump(int fd, char bank, FILE *outfile)
{
	char answer[100], command[100];
	int channel;
	char r_chan[4];
	int r_ctcss, r_freq;
	char r_trunk, r_delay, r_lockout, r_record;
	int x;

	set_mode(fd, M_MANUAL, 'Z');

	fprintf(outfile, "version %s\n", FFVERSION);

	channel = get_first_channel(bank);
	for (x = channel; channel <= x + 29; channel++)
	{
		bzero(answer, 100);
		bzero(command, 100);
		bzero(r_chan, 4);
		snprintf(command, 100, "PM%03d", channel);
		do_command(fd, command, answer);
		sscanf(answer, "C%3s F%8d T%c D%c L%c AF R%c N%2d", r_chan, &r_freq,
				&r_trunk, &r_delay, &r_lockout, &r_record, &r_ctcss);
		fprintf(outfile, "%d %c %c %c %c %d\n", r_freq, r_trunk, r_delay, r_lockout, r_record, r_ctcss);
	}

	fclose(outfile);  /* note this can close stdout if they didn't specify a
						 filename, but at this point we don't care */
	return(1);
}


/***********************************************************************
 * get the scanners current mode -- ie "trunking" "limit search" "manual"
 * Since we have different models of scanners to support, and each
 * model's raw mode returns a number that might mean one thing on one
 * model and another thing on another model, we have to figure out what
 * model they are using and return the correct mode for that model.
 * See commands.h to see what the mode identifiers mean.
 ***********************************************************************/
int get_mode(int fd)
{
	int raw_mode;
	int curr_mode;

	raw_mode = get_mode_raw(fd);
	switch (raw_mode)
	{
		case 0:
			curr_mode = M_CHAN_SCAN;
			break;
		case 1:
			curr_mode = M_MANUAL;
			break;
		case 2:
			curr_mode = M_LIMIT_SRCH;
			break;
		case 3:
			curr_mode = M_LIMIT_SRCH_HOLD;
			break;
		case 4:
			if (scannermodel == BC245XLT || scannermodel == PRO2052)
				curr_mode = M_SVC_SCAN;
			else if (scannermodel == BC895XLT || scannermodel == BC780XLT)
				curr_mode = M_WX_SCAN;
			break;
		case 5:
			if (scannermodel == BC895XLT)
			    curr_mode = M_WX_SCAN_HOLD;
			else
			    curr_mode = M_M_PGM_TRUNK;
			break;
		case 6:
			if (scannermodel == BC895XLT)
				curr_mode = M_M_PGM_TRUNK;
			else if (scannermodel == BC245XLT || scannermodel == PRO2052)
				curr_mode = M_E_PGM_TRUNK;
			else if (scannermodel == BC780XLT)
				curr_mode = M_UNKNOWN;
			break;
		case 7:
			if (scannermodel == BC895XLT)
				curr_mode = M_M_ID_SRCH;
			else if (scannermodel == BC245XLT || scannermodel == PRO2052)
				curr_mode = M_TRUNK_PGM;
			else if (scannermodel == BC780XLT)
				curr_mode = M_ROT_TUNE;
			break;
		case 8:
			if (scannermodel == BC895XLT)
				curr_mode = M_M_ID_SRCH_HOLD;
			else if (scannermodel == BC245XLT || scannermodel == PRO2052 || scannermodel == BC780XLT)
				curr_mode = M_M_ID_SRCH;
			break;
		case 9:
			if (scannermodel == BC895XLT)
				curr_mode = M_M_ID_SCAN;
			else if (scannermodel == BC245XLT || scannermodel == PRO2052 || scannermodel == BC780XLT)
				curr_mode = M_M_ID_SRCH_HOLD;
			break;
		case 10:
			if (scannermodel == BC895XLT)
				curr_mode = M_M_ID_MANUAL;
			else if (scannermodel == BC245XLT || scannermodel == PRO2052 || scannermodel == BC780XLT)
				curr_mode = M_M_ID_SCAN;
			break;
		case 11:
			if (scannermodel == BC895XLT)
				curr_mode = M_M_ID_LO;
			else if (scannermodel == BC245XLT || scannermodel == PRO2052 || scannermodel == BC780XLT)
				curr_mode = M_M_ID_MANUAL;
			break;
		case 12:
			if (scannermodel == BC895XLT)
				curr_mode = M_M_SRCH_CCHAN;
			else if (scannermodel == BC245XLT || scannermodel == PRO2052)
				curr_mode = M_M_ID_LO;
			else if (scannermodel == BC780XLT)
				curr_mode = M_UNKNOWN;
			break;
		case 13:
			if (scannermodel == BC895XLT)
				curr_mode = M_PGM_CTCSS;
			else if (scannermodel == BC245XLT || scannermodel == PRO2052 || scannermodel == BC780XLT)
				curr_mode = M_UNKNOWN;
			break;
		case 14:
			if (scannermodel == BC895XLT)
				curr_mode = M_WX_ALERT;
			else if (scannermodel == BC245XLT || scannermodel == PRO2052)
				curr_mode = M_E_ID_SRCH;
			else if (scannermodel == BC780XLT)
				curr_mode = M_UNKNOWN;
			break;
		case 15:
			if (scannermodel == BC895XLT)
				curr_mode = M_FREQ_SEND;
			else if (scannermodel == BC245XLT || scannermodel == PRO2052)
				curr_mode = M_E_ID_SRCH_HOLD;
			else if (scannermodel == BC780XLT)
				curr_mode = M_UNKNOWN;
			break;
		case 16:
			if (scannermodel == BC895XLT)
				curr_mode = M_AUTO_STORE;
			else if (scannermodel == BC245XLT || scannermodel == PRO2052)
				curr_mode = M_E_ID_SCAN;
			else if (scannermodel == BC780XLT)
				curr_mode = M_UNKNOWN;
			break;
		case 17:
			if (scannermodel == BC245XLT || scannermodel == PRO2052)
				curr_mode = M_E_ID_MANUAL;
			else if (scannermodel == BC895XLT || scannermodel == BC780XLT)
				curr_mode = M_ROT_TUNE;
			break;
		case 18:
			if (scannermodel == BC245XLT || scannermodel == PRO2052)
				curr_mode = M_E_ID_SCAN;
			else if (scannermodel == BC895XLT || scannermodel == BC780XLT)
				curr_mode = M_UNKNOWN;
			break;
		case 19:
			if (scannermodel == BC245XLT || scannermodel == PRO2052)
				curr_mode = M_E_SRCH_CCHAN;
			else if (scannermodel == BC895XLT || scannermodel == BC780XLT)
				curr_mode = M_UNKNOWN;
			break;
		case 20:
			if (scannermodel == BC245XLT || scannermodel == PRO2052)
				curr_mode = M_RF_TUNE;
			else if (scannermodel == BC895XLT || scannermodel == BC780XLT)
				curr_mode = M_UNKNOWN;
			break;
		default:
			curr_mode = M_UNKNOWN;
	}

	return(curr_mode);
}


/***********************************************************************
 * get "raw" mode
 ***********************************************************************/
int get_mode_raw(int fd)
{
	char answer[100];
	char modenum[3];
	int curr_mode;

	bzero(answer, 100);
	do_command(fd, "MD", answer);
	/* printf("%s\n", answer); */
	sscanf(answer, "MD%s", modenum);
	curr_mode = atoi(modenum);

	/* printf("raw mode: %d\n", curr_mode); */
	return(curr_mode);
}


/***********************************************************************
 * LCD backlight
 ***********************************************************************/
int light(int fd, int light_var)
{
	char answer[100];
	char blstat[3], alstat[3];

	bzero(answer, 100);

	if (light_var == -1)
	{
		do_command(fd, "LT", answer);

		if (strncmp(answer, "LTN", 3) == 0)
			strcpy(blstat, "ON");
		else if (strncmp(answer, "LTF", 3) == 0)
			strcpy(blstat, "OFF");
		else
			u_err("LT", answer);

		if (scannermodel == BC245XLT)
		{
			bzero(answer, 100);
			bzero(alstat, 3);
			do_command(fd, "AL", answer);

			if (strncmp(answer, "ALN", 3) == 0)
				strcpy(alstat, "ON");
			else if (strncmp(answer, "ALF", 3) == 0)
				strcpy(alstat, "OFF");
			else
				u_err("AL", answer);
		}
		else
			strcpy(alstat, "N/A");

		printf("Backlight: %s  Autolight: %s\n", blstat, alstat);
	}
	else if (light_var == 0)
	{
		do_command(fd, "LTF", answer);

		if (strncmp(answer, "OK", 2) == 0)
		{
			puts("Turned backlight OFF.");
			return(1);
		}
		else
			u_err("LTF", answer);
	}
	else if (light_var == 1)
	{
		do_command(fd, "LTN", answer);

		if (strncmp(answer, "OK", 2) == 0)
		{
			puts("Turned backlight ON.");
			return(1);
		}
		else
			u_err("LTN", answer);
	}
	else if (light_var == 2)
	{
		do_command(fd, "ALF", answer);

		if (strncmp(answer, "OK", 2) == 0)
		{
			puts("Turned autolight OFF.");
			return(1);
		}
		else
			u_err("ALF", answer);
	}
	else if (light_var == 3)
	{
		do_command(fd, "ALN", answer);

		if (strncmp(answer, "OK", 2) == 0)
		{
			puts("Turned autolight ON.");
			return(1);
		}
		else
			u_err("ALN", answer);
	}

	return(1);
}


/***********************************************************************
 * Used to specify the limits on the search command.
 ***********************************************************************/
int limit(int fd, int query, double freq1, double freq2)
{
	char answer[100], command[100];
	char lolimit[9];
	char hilimit[9];
	char freq_string[10];
	double tmp;

	if (query)
	{
		bzero(answer, 100);
		do_command(fd, "LL", answer);
		if (strncmp(answer, "LL", 2) != 0)
			u_err("LL", answer);
		sscanf(answer, "LL%8s", lolimit);

		bzero(answer, 100);
		do_command(fd, "LU", answer);
		if (strncmp(answer, "LU", 2) != 0)
			u_err("LU", answer);
		sscanf(answer, "LU%8s", hilimit);

		printf("Lower Limit: %.4f MHz\n", bc2human_freq(lolimit));
		printf("Upper Limit: %.4f MHz\n", bc2human_freq(hilimit));

		return(1);
	}

	if (withinlimits(freq1) == 1 && withinlimits(freq2) == 1)  /* check that freqs are within coverage range */
	{
		/* in case they entered them highest first */
		if (freq1 > freq2) {
			tmp = freq1;
			freq1 = freq2;
			freq2 = tmp;
		}

		snprintf(freq_string, 9, "%f", freq1);
		human2bc_freq(freq_string);
		snprintf(command, 100, "LL%s", freq_string);
		bzero(answer, 100);
		do_command(fd, command, answer);
		printf("New Lower Limit: %.4f MHz\n", freq1);

		snprintf(freq_string, 9, "%f", freq2);
		human2bc_freq(freq_string);
		snprintf(command, 100, "LU%s", freq_string);
		bzero(answer, 100);
		do_command(fd, command, answer);
		printf("New Upper Limit: %.4f MHz\n", freq2);

		return(1);
	}
	else
	{
		return(-1);  /* mhz out of range */
	}

	return(1);
}


/***********************************************************************
 * Upload a bank from a file to the scanner
 ***********************************************************************/
int load(int fd, char bank, FILE *infile)
{
	char answer[100], command[100], errstring[100];
	int channel, ffversion;
	int r_ctcss, r_freq;
	char r_trunk, r_delay, r_lockout, r_record;
	int x;

	set_mode(fd, M_MANUAL, 'Z');

	fscanf(infile, "version %d", &ffversion);
	if (ffversion != 1)
	{
		snprintf(errstring, 100, "Sctl can only read dumped file format version %s.", FFVERSION);
		complain(errstring);
		return(-1);
	}

	channel = get_first_channel(bank);
	for (x = channel; channel <= x + 29; channel++)
	{
		fscanf(infile, "%d %c %c %c %c %d", &r_freq, &r_trunk, &r_delay, &r_lockout, &r_record, &r_ctcss);
		/* too bad we can't just upload them as easily as we downloaded them!
		 * printf("%03d %08d %c %c %c %c %02d\n", channel, r_freq, r_trunk, r_delay, r_lockout, r_record, r_ctcss); */

		/* set frequency */
		bzero(answer, 100);
		bzero(command, 100);
		snprintf(command, 100, "PM%03d %08d", channel, r_freq);
		do_command(fd, command, answer);
		/* set trunk */

		/* set delay */
		if (r_delay == 'F')
			delay(fd, 0);
		else
			delay(fd, 1);
		/* set lockout */
		if (r_lockout == 'F')
			lockout(fd, 0, 0);
		else
			lockout(fd, 1, 0);
		/* set autorecord */

		/* set ctcss */

		/* refresh the display and change to first memory location in bank */
		snprintf(command, 100, "MA%03d", get_first_channel(bank));
		do_command(fd, command, answer);
	}

	fclose(infile);  /* note this can close stdin if they didn't specify a filename,
						but at this point we don't care */
	return(1);
}


/***********************************************************************
 ***********************************************************************/
int lockout(int fd, int lockout_var, int force)
{
	char answer[100];

	bzero(answer, 100);

	if (lockout_var == -1)
	{
		if (force)
		{
			set_mode(fd, M_MANUAL, 'Z');
		}

		do_command(fd, "LO", answer);

		if (strncmp(answer, "LON", 3) == 0)
		{
			puts("Lockout function is ON for this channel.");
			return(1);
		}
		else if (strncmp(answer, "LOF", 3) == 0)
		{
			puts("Lockout function is OFF for this channel.");
			return(1);
		}
		else
			u_err("LO", answer);
	}
	else if (lockout_var == 0)
	{
		do_command(fd, "LOF", answer);

		if (strncmp(answer, "OK", 2) == 0)
		{
			puts("Set lockout OFF for this channel.");
			return(1);
		}
		else
			u_err("LOF", answer);
	}
	else if (lockout_var == 1)
	{
		do_command(fd, "LON", answer);

		if (strncmp(answer, "OK", 2) == 0)
		{
			puts("Set lockout ON for this channel.");
			return(1);
		}
		else
			u_err("LON", answer);
	}

	return(1);
}


/***********************************************************************
 * used for reading/writing frequencies in the scanner's memory
 ***********************************************************************/
int mem(int fd, int channel, double freq, int force, int set)
{
	char answer[100], command[100], errorstr[100];
	double human_freq;
	char r_chan[4], r_freq[9], freq_string[10];
	char human_trunk[4], human_delay[4], human_lockout[4], human_record[4];
	int r_ctcss;
	char r_trunk, r_delay, r_lockout, r_record;

	int curr_mode = get_mode(fd); 

	bzero(answer, 100);
	bzero(command, 100);
	bzero(errorstr, 100);
	bzero(r_chan, 4);
	bzero(r_freq, 9);
	bzero(freq_string, 10);
	bzero(human_trunk, 4);
	bzero(human_delay, 4);
	bzero(human_lockout, 4);
	bzero(human_record, 4);

	if (set == 0)
	{
		snprintf(command, 100, "PM%03d", channel);
		do_command(fd, command, answer);

		sscanf(answer, "C%3s F%8s T%c D%c L%c AF R%c N%2i", r_chan, r_freq,
			&r_trunk, &r_delay, &r_lockout, &r_record, &r_ctcss);

		human_freq = bc2human_freq(r_freq);
		bc2human_onoff(r_trunk, human_trunk);
		bc2human_onoff(r_delay, human_delay);
		bc2human_onoff(r_lockout, human_lockout);
		bc2human_onoff(r_record, human_record);

		printf("Memory %i: %.4f  Trunk: %s  Delay: %s  LO: %s  AutoRec: %s  CTCSS: %d\n",
			atoi(r_chan), human_freq, human_trunk, human_delay, human_lockout,
			human_record, r_ctcss);
		return(1);
	}

	if (curr_mode != M_MANUAL && !force)
	{
		complain("Radio is not in manual mode.  Use -f to override.");
		return(-1);
	}
	else if (curr_mode != M_MANUAL && force)
	{
		set_mode(fd, M_MANUAL, 'Z'); 
	}

	if (freq == -1) /* aka, don't set the freq */
	{
		if (channel != -1)
		{
			snprintf(command, 100, "MA%03d", channel);
			do_command(fd, command, answer);
			printf("Set current memory channel to %03d.\n", channel);
		}
		else
			do_command(fd, "MA", answer);

		sscanf(answer, "C%3s F%8s T%c D%c L%c AF R%c N%2i", r_chan, r_freq,
			&r_trunk, &r_delay, &r_lockout, &r_record, &r_ctcss);

		human_freq = bc2human_freq(r_freq);
		bc2human_onoff(r_trunk, human_trunk);
		bc2human_onoff(r_delay, human_delay);
		bc2human_onoff(r_lockout, human_lockout);
		bc2human_onoff(r_record, human_record);

		printf("Memory %i: %.4f  Trunk: %s  Delay: %s  LO: %s  AutoRec: %s  CTCSS: %d\n",
			atoi(r_chan), human_freq, human_trunk, human_delay, human_lockout,
			human_record, r_ctcss);
		return(1);
	}
	else
	{
		if (freq != 0)                          /* check for bad frequencies */
			if (withinlimits(freq) == -1)
				return(-1);

		snprintf(freq_string, 9, "%f", freq);
		human2bc_freq(freq_string);

		bzero(answer, 100);
		bzero(command, 100);
		snprintf(command, 100, "PM%03d %s", channel, freq_string);
		do_command(fd, command, answer);
		
		if(sscanf(answer, "C%3s F%8s", r_chan, r_freq))
		{
			printf("Memory channel %i set to %.4f\n", channel, freq);
			return(1);
		}
		else
			u_err(command, answer);
	}

	return(1);
}


/***********************************************************************
 * Mute setting
 ***********************************************************************/
int mute(int fd, int mute_var)
{
	char answer[100];

	bzero(answer, 100);

	if (mute_var == 0)
	{
		do_command(fd, "MUF", answer);

		if (strncmp(answer, "OK", 2) == 0)
		{
			puts("Set mute OFF.");
			return(1);
		}
		else
			u_err("MUF", answer);
	}
	else if (mute_var == 1)
	{
		do_command(fd, "MUN", answer);

		if (strncmp(answer, "OK", 2) == 0)
		{
			puts("Set mute to ALL.");
			return(1);
		}
		else
			u_err("MUN", answer);
	}
	else if (mute_var == 2)
	{
		do_command(fd, "MUA", answer);

		if (strncmp(answer, "OK", 2) == 0)
		{
			puts("Set mute to NORMAL.");
			return(1);
		}
		else
			u_err("MUA", answer);
	}
	else if (mute_var == -1)
	{
		do_command(fd, "MU", answer);

		if (strncmp(answer, "MUF", 3) == 0)
		{
			puts("Speaker muting is OFF.");
			return(1);
		}
		else if (strncmp(answer, "MUN", 3) == 0)
		{
			puts("Speaker muting is ON.");
			return(1);
		}
		else if (strncmp(answer, "MUA", 3) == 0)
		{
			puts("Speaker muting is NORMAL.");
			return(1);
		}
		else
			u_err("MU", answer);
	}

	return(1);
}


/***********************************************************************
 * press button KEY on scanner
 ***********************************************************************/
int press_key(int fd, char *key)
{
	char answer[100];

	bzero(answer, 100);
	do_command(fd, key, answer);

	if (strncmp(answer, "OK", 2) != 0)
		u_err(key, answer);

	return(1);
}


/***********************************************************************
 * conventional scan
 ***********************************************************************/
int scan(int fd, char *banks)
{
	if (set_scan_banks(fd, banks) != 1)
	{
		return(-1);
	}
	
	set_mode(fd, M_CHAN_SCAN, 'Z');
	return(1);
}


/***********************************************************************
 * scan trunked banks
 ***********************************************************************/
int scan_trunk(int fd, char *banks)
{
	char answer[100];
	char command[100];

	set_mode(fd, M_MANUAL, 'Z');

	if (scannermodel == BC895XLT)
	{
		/* this only seems to be true on the 895?? */
		/* I am able to scan multiple trunked banks on my 245 */
		if (banks[1] != '\0')
		{
			complain("Can't trunk scan more than one bank on this radio.");
			return(-1);
		}
	}

	press_key(fd, K_TRUNK);

	if (scannermodel != BC895XLT)
	/* the 245 doesn't have a KEY21 --
	 * the regular set_scan_banks function does the job on the 245 */
	{
		if (set_scan_banks(fd, banks) == -1)
			return(-1);
	}
	else
	{
		printf("Setting radio to scan bank %s.\n", banks);
		bzero(answer, 100);
		snprintf(command, 100, "KEY21 %c", banks[0]);
		do_command(fd, command, answer);
		if (strncmp(answer, "OK", 2) != 0)
			u_err(command, answer);
	}

	press_key(fd, K_SCAN);

	return(1);
}


/***********************************************************************
 * search for a frequency within a certain range
 ***********************************************************************/
int search(int fd)
{
	press_key(fd, K_SRC);

	return(1);
}


/***********************************************************************
 ***********************************************************************/
int set_mode(int fd, int new_mode, char bank)
{
	int old_mode = get_mode(fd);
	char keystroke[10];

	bzero(keystroke, 10);

	/* switch to mode 01 first, then go to our new mode */
	/* except "hold modes" which need to stay in their mode to work */
	/* assumption currently seems to be that if we're going from a
	 * non-trunk mode to a non-trunk mode, a trunk mode to a trunk
	 * mode, or a trunk mode to a non-trunk mode, it'll work, but there 
	 * seems to be no provision for going from non-trunk to trunk.
	 */
	switch (old_mode)
	{
		case M_LIMIT_SRCH:
			if (new_mode == M_LIMIT_SRCH_HOLD)
				break; /* don't fallthrough if we need to hold this mode*/
			/* otherwise fallthrough to hitting K_MANUAL */

		case M_WX_SCAN:
			if (new_mode == M_WX_SCAN_HOLD)
				break; /* don't fallthrough if we need to hold this mode*/
			/* otherwise fallthrough to hitting K_MANUAL */

		case M_MANUAL:
		case M_CHAN_SCAN:
		case M_LIMIT_SRCH_HOLD:
		case M_WX_SCAN_HOLD:
		case M_PGM_CTCSS:
		case M_WX_ALERT:
		case M_ROT_TUNE:
			if (new_mode == M_M_PGM_TRUNK || new_mode == M_M_ID_SRCH_HOLD
				|| new_mode == M_M_ID_SCAN || new_mode == M_M_ID_MANUAL ||
				new_mode == M_M_ID_LO || new_mode == M_M_SRCH_CCHAN)
			{
				/* ie, we're going into a trunk mode from a non-trunk */
				if (bank < 'A' || bank > 'J')
				{
					complain("Must specify a bank between A and J when switching from normal mode to trunk mode.");
					return(-1);
				}

				press_key(fd, K_MANUAL);
				press_key(fd, K_TRUNK);
				snprintf(keystroke, 10, "KEY21 %c", bank);
				press_key(fd, keystroke);
				sleep(2);
			}
			else
			{
				/* otherwise going into a non-trunk */
				press_key(fd, K_MANUAL);
			}
			break;

		case M_M_ID_SRCH:
			if (new_mode == M_M_ID_SRCH_HOLD)
				break; /* don't fallthrough if we need to hold this mode*/

		case M_M_PGM_TRUNK:
		case M_M_ID_SRCH_HOLD:
		case M_M_ID_SCAN:
		case M_M_ID_MANUAL:
		case M_M_ID_LO:
		case M_M_SRCH_CCHAN:
			/* if we're going to another trunk mode, do nothing */
			if (new_mode == M_M_PGM_TRUNK || new_mode == M_M_ID_SRCH_HOLD
				|| new_mode == M_M_ID_SCAN || new_mode == M_M_ID_MANUAL ||
				new_mode == M_M_ID_LO || new_mode == M_M_SRCH_CCHAN)
			{
				if (bank == 'Z')
				{
					/* don't change the bank if we get Z, since that's
					 * the "do nothing" value */
					break;
				}
				else
				{
					if (bank < 'A' || bank > 'J')
					{
						complain("Can't use bank numbers other than A through J.");
						return(-1);
					}

					/* left off here.  this sequence is wrong. */
					press_key(fd, K_TRUNK);
					press_key(fd, K_MANUAL);
					press_key(fd, K_TRUNK);
					snprintf(keystroke, 10, "KEY21 %c", bank);
					press_key(fd, keystroke);
					sleep(2);
				}
			}
			else
			{
				/* get us back out of trunk mode */
				press_key(fd, K_TRUNK);
				break;
			}

		default:
			break;
	}

	switch (new_mode)
	{
		case M_CHAN_SCAN:
			press_key(fd, K_SCAN);
			break;

		case M_MANUAL:
			break;

		case M_LIMIT_SRCH:
			press_key(fd, K_SRC);
			break;

		case M_LIMIT_SRCH_HOLD: /* already in search mode */
			press_key(fd, K_HOLD_UP);
			break;

		case M_WX_SCAN:
			press_key(fd, K_WX);
			break;

		case M_WX_SCAN_HOLD: /* already in search mode */
			press_key(fd, K_HOLD_UP);
			break;

		case M_M_PGM_TRUNK:
			/* no clue, skip for now */
			complain("This mode is not possible in computer control mode.");
			break;

		case M_M_ID_SRCH:
			/* press_key(fd, K_SRC); */
			break;

		case M_M_ID_SRCH_HOLD: /* already in search mode */
			press_key(fd, K_HOLD_UP);
			break;
		
		case M_M_ID_SCAN:
			press_key(fd, K_SCAN);
			break;

		case M_M_ID_MANUAL:
			press_key(fd, K_MANUAL);
			break;

		case M_M_ID_LO:
			press_key(fd, K_LO);
			break;

		case M_M_SRCH_CCHAN:
			complain("This mode cannot be set.");
			break;

		case M_PGM_CTCSS:
			complain("This mode is not yet supported.");
			break;

		case M_WX_ALERT:
			press_key(fd, K_WX);
			press_key(fd, K_ALERT_REMOTE);
			break;

		case M_FREQ_SEND:
			complain("This mode is not yet supported.");
			break;

		case M_AUTO_STORE:
			/* to do this right, we have to know which banks to store,
			 * so let's skip it for now. */
			complain("This mode is not yet supported.");
			break;

		case M_ROT_TUNE:
			complain("This mode may not be set in computer control mode.");
			break;
	}

	return(1);
}


/***********************************************************************
 * activate some banks so we can scan them
 ***********************************************************************/
int set_scan_banks(int fd, char *banks)
{
	char command[100], answer[100];
	char likely_answer[15], likely_answer2[15];

	bzero(answer, 100);
	bzero(command, 100);
	bzero(likely_answer, 15);
	bzero(likely_answer2, 15);

	if (scannermodel == BC895XLT)
		snprintf(command, 100, "SB %s\r", banks);
	else
		snprintf(command, 100, "SB %s", banks);

	snprintf(likely_answer, 15, "SB %s\n", banks);	/* scanner can return either */
	snprintf(likely_answer2, 15, "SB %s\r", banks); /* seems to vary based on whether
													 * or not it is in trunking mode? */
	printf("Set radio to scan banks %s.\n", banks);
	do_command(fd, command, answer);

	if (strcmp(answer, likely_answer) != 0 && strcmp(answer, likely_answer2) != 0)
		u_err(command, answer);

	return(1);
}


/***********************************************************************
 * set/query talkgroup
 ***********************************************************************/
int talkgroup(int fd, char *memloc, int groupid, char bank)
{
	char answer[100], command[100], l_memloc[3], errorstr[100], l_gid[15];
	int l_groupid, i;

	bzero(answer, 100);
	bzero(command, 100);
	bzero(l_memloc, 3);
	bzero(errorstr, 100);
	bzero(l_gid, 15);

	/* for some reason, IC is switching us to manual mode.  must fix. */
	if (groupid == -1)
	{
		set_mode(fd, M_M_ID_MANUAL, bank);

		if (memloc[0] == '\0')
		{
			snprintf(command, 99, "IC");
		}
		else
		{
			snprintf(command, 99, "IC %s", memloc);
			printf("Setting radio to talkgroup memory %s\n", memloc);
		}

		do_command(fd, command, answer);
	}
	else
	{
		set_mode(fd, M_M_ID_MANUAL, bank);
		snprintf(command, 99, "IC %s", memloc);
		do_command(fd, command, answer);

		if (sscanf(answer, "IC %2s %d", l_memloc, &l_groupid) != 2)
		{
			if (sscanf(answer, "IC %2s ------", l_memloc) == 1)
			{
				/* all is well, the memory just hasn't been set */
			}
			else
			{
				if (strncmp(answer, "NG", 2) == 0)
				{
					complain("The radio must be in trunk mode if the bank is omitted.");
					return(-1);
				}
				snprintf(errorstr, 99, "Didn't get both mem and groupid: %s", answer);
				complain(errorstr);
				return(-1);
			}
		}

		snprintf(l_gid, 14, "%d", groupid);

		/* press each number key in turn */
		for (i=0; l_gid[i] != '\0'; i++)
		{
			snprintf(command, 99, "KEY02 %c", l_gid[i]);
			if (press_key(fd, command) == -1)
			{
				snprintf(errorstr, 99, "Couldn't press number %c, giving up.", 
					l_gid[i]);
				complain(errorstr);
				return(-1);
			}
		}

		if (press_key(fd, K_ENTER) == -1)
		{
			complain("Couldn't press enter, giving up.");
			return(-1);
		}

		sleep(1);
		bzero(answer, 99);
		snprintf(command, 99, "IC %s", memloc);
		printf("Setting radio to talkgroup memory %s\n", l_memloc);
		do_command(fd, command, answer);
	}

	if (sscanf(answer, "IC %2s %d", l_memloc, &l_groupid) != 2)
	{
		if (sscanf(answer, "IC %2s ------", l_memloc) == 1)
		{
			if (groupid == -1)
			{
				printf("Talkgroup memory %s is empty (no value set)\n", 
					l_memloc);
			}
			else
			{
				printf("Talkgroup memory %s has been set to empty\n", 
					l_memloc);
			}
			return(1);
		}

		if (strncmp(answer, "NG", 2) == 0)
		{
			complain("Radio must be paused in a trunking mode to use this function.");
			return(-1);
		}
		snprintf(errorstr, 99, "Didn't get both mem and groupid: %s", answer);
		complain(errorstr);
		return(-1);
	}

	if (groupid == -1)
	{
		printf("Talkgroup in memory %s is %d\n", l_memloc, l_groupid);
	}
	else
	{
		printf("Talkgroup in memory %s has been set to %d\n", l_memloc, 
			l_groupid);
	}
		
	return(1);
}


/***********************************************************************
 * trying to use the regular talkgroup command to list all memories was
 * painfully slow, so i added this one.
 ***********************************************************************/
int talkgroup_all(int fd, char bank)
{
	char answer[100], command[100], l_memloc[3], errorstr[100], l_gid[15];
	char b;
	int l_groupid, i;

	set_mode(fd, M_M_ID_MANUAL, bank);

	for (b='A'; b<'F'; b++)
	{
		for (i=0; i<10; i++)
		{
			bzero(answer, 100);
			bzero(command, 100);
			bzero(l_memloc, 3);
			bzero(errorstr, 100);
			bzero(l_gid, 15);

			snprintf(command, 99, "IC %c%d", b, i);

			do_command(fd, command, answer);

			if (sscanf(answer, "IC %2s %d", l_memloc, &l_groupid) != 2)
			{
				if (sscanf(answer, "IC %2s ------", l_memloc) == 1)
				{
					printf("Talkgroup memory %s is empty (no value set).\n", 
						l_memloc);
					continue;
				}

				if (strncmp(answer, "NG", 2) == 0)
				{
					complain("Radio must be paused in a trunking mode to use this function.");
					return(-1);
				}
				snprintf(errorstr, 99, "Didn't get both mem and groupid: %s", answer);
				complain(errorstr);
				return(-1);
			}

			printf("Talkgroup in memory %s is %d\n", l_memloc, l_groupid);
		}
	}
		
	return(1);
}


/***********************************************************************
 ***********************************************************************/
int version(int fd)
{
	char answer[100];

	printf("Sctl %s\n", VERSION);
	printf("Copyright (c) 2001\n");
	printf("Sctl comes with NO WARRANTY, to the extent permitted by law.  You may\n");
	printf("redistribute copies of Sctl under the terms of the GNU General Public License.\n");
	printf("For more information on these matters, see the file named LICENSE.\n\n");

	printf("Scanner Model: %d\n", scannermodel);
	if (scannermodel != BC895XLT)
	/* You might wonder why we don't just run this command to determine what model of
	 * scanner the user is using, and automatically create a config file.  Well, the
	 * bc895xlt does not support this command, so it wouldn't work (it would probably
	 * return ERR).  Technically we could just assume that anything that returns "ERR"
	 * is an 895 but I don't trust Uniden to be that consistent ;)
	 * In the future it might be cool to add the autoconfig functionality though.  */
	{
		bzero(answer, 100);
		do_command(fd, "SI", answer);
		printf("Scanner Sysinfo: %s", answer);
		bzero(answer, 100);
		do_command(fd, "VR", answer);
		printf("Scanner Firmware: %s", answer);
	}

	return(1);
}


/***********************************************************************
 * this function provides a running account of talkgroups, reporting
 * time and date, talkgroup, frequency and signal strength
 ***********************************************************************/
int id_report(int fd)
{
	char answer[100];
	int size, tg_id;
	char event, event_name[6], c_temp[20], c_temp2[20];
	struct tm *curr_time;
	char timestamp[50];
	time_t epoch;
	double d_temp;
	int mode;

	bzero(c_temp, 20);
	bzero(c_temp2, 20);
	bzero(answer, 100);
	bzero(event_name, 6);
	bzero(timestamp, 50);

	mode = get_mode(fd);
	if (mode != M_M_ID_SRCH && mode != M_M_ID_SCAN && mode != M_M_ID_MANUAL &&
		mode != M_M_SRCH_CCHAN && mode != M_E_ID_SRCH && mode != M_E_ID_SCAN &&
		mode != M_E_ID_MANUAL && mode != M_E_SRCH_CCHAN)
	{
		puts("--- Not in a trunking mode, this will probably be useless\n");
	}

	do_command(fd, "IDN", answer);

	if (strncmp(answer, "OK", 2) == 0)
	{
		puts("ID report mode enabled\n");
	}
	else if (strncmp(answer, "NG", 2) == 0)
	{
		puts("Unable to enter ID report mode\n");
		return(-1);
	}

	bzero(answer, 100);

	while ((size = read_serial(fd, answer)))
	{
		epoch = time(NULL);
		curr_time = localtime(&epoch);
		sprintf(timestamp, "%02d/%02d %02d:%02d:%02d", curr_time->tm_mon + 1, 
			curr_time->tm_mday, curr_time->tm_hour, curr_time->tm_min,
			curr_time->tm_sec);

		sscanf(answer, "ID %c %d", &event, &tg_id);

		if (strncmp(&event, "S", 1) == 0)
		{
			bzero(event_name, 6);
			strncpy(event_name, "start", 5);

			bzero(answer, 100);
			do_command(fd, "SG", answer);
			sscanf(answer, "S%3s F%8s", c_temp, c_temp2);
			d_temp = bc2human_freq(c_temp2);

			if (scannermodel == BC895XLT)
			{
				printf("%s %s ID: %06d Freq: %.4f S: %s  ", timestamp, 
					event_name, tg_id, d_temp, c_temp);

				print_s_meter(c_temp);

				putchar('\n');
			}
			else
			{
				printf("%s %s ID: %06d Freq: %.4f S: %s\n", timestamp, 
					event_name, tg_id, d_temp, c_temp);
			}
		}
		else if (strncmp(&event, "E", 1) == 0)
		{
			bzero(event_name, 6);
			strncpy(event_name, "end  ", 5);

			printf("%s %s ID: %06d\n", timestamp, event_name, 
				tg_id);
		}

		bzero(answer, 100);
	}

	return(1);
}


/***********************************************************************
 * this function provides an ongoing report on received audio in 
 * conventional scan mode, very much the conventional equivalent to
 * id_report.  uses QU to get squelch changes, then gets frequency info
 * via SG.  sets the scan mode.
 ***********************************************************************/
int scan_report(int fd, char *banks)
{
	char answer[100], event, event_name[10], c_temp[20], c_temp2[20];
	char timestamp[50];
	int size;
	double d_temp;
	struct tm *curr_time;
	time_t epoch;

	bzero(answer, 100);
	bzero(event_name, 10);
	bzero(timestamp, 50);
	bzero(c_temp, 20);
	bzero(c_temp2, 20);
	bzero(timestamp, 50);

	if (scan(fd, banks) != 1)
	{
		puts("Unable to set scanning mode\n");
		return(-1);
	}
	
	do_command(fd, "QUN", answer);

	if (strncmp(answer, "OK", 2) == 0)
	{
		printf("Scanning banks %s\n", banks);
	}
	else if (strncmp(answer, "NG", 2) == 0)
	{
		puts("Unable to enter scan report mode\n");
		return(-1);
	}

	bzero(answer, 100);

	while ((size = read_serial(fd, answer)))
	{
		epoch = time(NULL);
		curr_time = localtime(&epoch);
		sprintf(timestamp, "%02d/%02d %02d:%02d:%02d", curr_time->tm_mon + 1, 
			curr_time->tm_mday, curr_time->tm_hour, curr_time->tm_min,
			curr_time->tm_sec);

		sscanf(answer, "%c", &event);

		if (event == '+')
		{
			bzero(event_name, 6);
			strncpy(event_name, "start", 5);

			bzero(answer, 100);
			do_command(fd, "SG", answer);
			sscanf(answer, "S%3s F%8s", c_temp, c_temp2);
			d_temp = bc2human_freq(c_temp2);

			if (scannermodel == BC895XLT)
			{
				printf("%s %s Freq: %.4f S: %s  ", timestamp, event_name, 
					d_temp, c_temp);

				print_s_meter(c_temp);

				putchar('\n');
			}
			else
			{
				printf("%s %s Freq: %.4f S: %s\n", timestamp, 
					event_name, d_temp, c_temp);
			}
		}
		else if (event == '-')
		{
			bzero(event_name, 6);
			strncpy(event_name, "end  ", 5);

			printf("%s %s\n", timestamp, event_name);
		}

		bzero(answer, 100);
	}

	return(1);
}
